/* -*- Mode: C; tab-width: 4; -*- */
/* npdefpluginpro.c
 *
 * DefaultPluginPro
 *
 * Copyright (c) 2002 Jason Summers <jason1@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 * See the files readme.txt and COPYING.txt for more information.
 *
 */

#define DPPVERS    "1.0.0"

#ifdef UNICODE
#error
#endif

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "resource.h"
#include "npapidefs.h"

#define IDBASE           46000
#define ID_SAVEAS        (IDBASE+0)
#define ID_COPYURL       (IDBASE+1)
#define ID_VIEWDIRECT    (IDBASE+2)
#define ID_OPENNEWWIN    (IDBASE+3)
#define ID_PLUGINSEARCH  (IDBASE+4)
#define ID_PLUGINSPAGE   (IDBASE+5)
#define ID_ABOUT         (IDBASE+6)
#define ID_LOADDOC       (IDBASE+7)
#define ID_STOPLOAD      (IDBASE+8)
#define ID_PROPERTIES    (IDBASE+9)
#define ID_REFRESHPLUGINS (IDBASE+10)
#define ID_AUTODOWNLOAD  (IDBASE+11)
#define ID_MENUCANCEL    (IDBASE+20)
#define ID_AUTODLYES     (IDBASE+30)
#define ID_AUTODLNO      (IDBASE+31)
#define ID_AUTODLDEF     (IDBASE+32)

#define ST_INIT        0
#define ST_NOTLOADING  1
#define ST_STOPPED     2
#define ST_LOADING     3
#define ST_LOADED      4

/* plugin instance-specific data */
typedef struct plugininstance_struct
{
	NPWindow*  fWindow;
	uint16     fMode;
	HWND       fhWnd;
	WNDPROC    fDefaultWindowProc;
	NPP        instance;

	char *url;         /* the url of the stream */
	char *pluginspage;
	char *mimetype;

	int status;      /* is the document finished loading? (ST_...) */
	int bytesread;
	int bytestotal;   /* expected size of datafile, or (-1) if unknown */
	int autodownload;   // the default autodownload setting
	int thistype_autodl;  /* autodownload this type? */
	int bogusmimetype;

	HBRUSH bkgdbrush1, bkgdbrush2, bkgdbrush3;
	HPEN outlinepen;
	COLORREF bgcolor1,bgcolor2, bgcolor3;
	HCURSOR cursor;

	NPStream *stream_id;
	int tempfilestatus; // 0=none, 1=created and opened, 2=closed and available
	HANDLE tempfile;
	char *tempfilename;
	int abortflag;
	int forcedownload;
	char *pluginsearchpage;    // NULL means use default
} PluginInstance;


/* global variables */
static const char* gInstanceLookupString = "pdata";
static char g_RegKey[] = "SOFTWARE\\Generic\\DefaultPluginPro";
static char g_DefPluginSearchPage[] = "http://cgi.netscape.com/cgi-bin/plug-in_finder.cgi";
static HMODULE g_hInst = NULL;
static NPNetscapeFuncs* g_pNavigatorFuncs = NULL;
static NPPluginFuncs* g_pluginFuncs = NULL;

/* function prototypes */
static LRESULT CALLBACK DlgProcProperties(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK DlgProcAbout(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK PluginWindowProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);

/* ----------------------------------------------------------------------- */

// the only NPN_ function we need:

static NPError NPN_GetURL(NPP instance, const char *url, const char *target)
{
    return g_pNavigatorFuncs->geturl(instance, url, target);
}

/* ----------------------------------------------------------------------- */

static void warn(PluginInstance *This, char *fmt, ...)
{
	va_list ap;
	char buf[2048];
	HWND hwnd;

	va_start(ap, fmt);
	vsprintf(buf,fmt, ap);
	va_end(ap);

	if(This) hwnd=This->fhWnd;
	else hwnd=NULL;

	MessageBox(hwnd,buf,"DefaultPluginPro",MB_OK|MB_ICONWARNING);
}

static void RepaintWindow(PluginInstance *This)
{
	if(This->fhWnd) {
		InvalidateRect(This->fhWnd,NULL,FALSE);
	}
}

static void ReadSettings(PluginInstance *This)
{
	HKEY key;
	DWORD datasize, datatype;
	DWORD tmpdword;
	LONG rv;
	char buf[1024];

	This->pluginsearchpage = NULL;
	This->autodownload = 0;

	if(RegOpenKeyEx(HKEY_CURRENT_USER,g_RegKey,0,KEY_READ,&key)
		!=ERROR_SUCCESS)
	{
		return;
	}

	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"autodownload",NULL,&datatype,(LPBYTE)&tmpdword,&datasize);
	if(rv==ERROR_SUCCESS && datatype==REG_DWORD) {
		This->autodownload = tmpdword;
	}

	datasize=1024;
	rv=RegQueryValueEx(key,"searchpage",NULL,&datatype,(LPBYTE)buf,&datasize);
	if(rv==ERROR_SUCCESS && datatype==REG_SZ) {
		// only store if not the default
		if(lstrcmp(buf,g_DefPluginSearchPage)) {
			This->pluginsearchpage = _strdup(buf);
		}
	}

	RegCloseKey(key);
}

static int WriteSetting(PluginInstance *This,char *subkey, char *name,
						DWORD type, BYTE *value, int length)
{
	DWORD disp;
	HKEY key;
	LONG rv;
	char buf[200];

	if(!name) return 0;
	if(lstrlen(name)<1) return 0;

	if(subkey) {
		sprintf(buf,"%s\\%s",g_RegKey,subkey);
	}
	else {
		sprintf(buf,"%s",g_RegKey);
	}

	rv=RegCreateKeyEx(HKEY_CURRENT_USER,buf,0,NULL,
		REG_OPTION_NON_VOLATILE,KEY_WRITE,NULL,&key,&disp);
	if(rv!=ERROR_SUCCESS) return 0;
	rv=RegSetValueEx(key,name,0,type,value,length);
	RegCloseKey(key);
	return (rv==ERROR_SUCCESS);
}

static int WriteSetting_Int(PluginInstance *This,char *subkey, char *name, int value)
{
	return WriteSetting(This,subkey,name,REG_DWORD,(BYTE*)&value,sizeof(DWORD));
}

// 0== use default, 1==never download, 2==always download
static int GetTypeDLPolicy(PluginInstance *This, char *mimetype)
{
	HKEY key;
	DWORD datasize, datatype;
	DWORD tmpdword;
	LONG rv;
	int retval;
	char buf[200];

	sprintf(buf,"%s\\types",g_RegKey);
	retval = 0;
	if(!mimetype) return retval;

	if(RegOpenKeyEx(HKEY_CURRENT_USER,buf,0,KEY_READ,&key)
		!=ERROR_SUCCESS)
	{
		return retval;
	}

	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,mimetype,NULL,&datatype,(LPBYTE)&tmpdword,&datasize);
	if(rv==ERROR_SUCCESS && datatype==REG_DWORD) {
		retval=tmpdword;
	}

	RegCloseKey(key);
	return retval;
}

// Find and record download policy for this type,
// and detect types that we shouldn't bother to try a search for
static void CheckMIMEType(PluginInstance *This)
{
	static char *badtypes[] = {"text/plain","application/octet-stream",
		"application/x-unknown-content-type","",NULL};
	int i;

	This->bogusmimetype=0;
	This->thistype_autodl=0; 
	if(!This->mimetype) return;
	for(i=0;badtypes[i];i++) {
		if(!lstrcmp(This->mimetype,badtypes[i])) {
			This->bogusmimetype=1;
		}
	}

	This->thistype_autodl = GetTypeDLPolicy(This, This->mimetype);
}

/* ----------------------------------------------------------------------- */

/* Once-per-instance initialization */
static NPError NPP_New(NPMIMEType pluginType,NPP instance,uint16 mode,
		int16 argc,char* argn[],char* argv[],NPSavedData* saved)
{
	PluginInstance* This;
	int i;

	if (!instance) {
		return NPERR_INVALID_INSTANCE_ERROR;
	}
	instance->pdata = calloc(sizeof(PluginInstance),1);
	This = (PluginInstance*)instance->pdata;
	if (!This) {
	    return NPERR_OUT_OF_MEMORY_ERROR;
	}

	This->fWindow = NULL;
	This->fMode = mode;

	This->fhWnd = NULL;
	This->fDefaultWindowProc = NULL;
	This->instance = instance;

	ReadSettings(This);

	This->status=ST_INIT;
	This->abortflag=0;
	This->forcedownload=0; // download even if it's against policy

	This->bgcolor1 = RGB(240,200,150);
	This->bgcolor2 = RGB(255,224,192);
	This->bgcolor3 = RGB(192,255,192);
	This->outlinepen=CreatePen(PS_DOT,0,RGB(255,0,0));
	This->bkgdbrush1=CreateSolidBrush(This->bgcolor1);
	This->bkgdbrush2=CreateSolidBrush(This->bgcolor2);
	This->bkgdbrush3=CreateSolidBrush(This->bgcolor3);
	This->cursor=LoadCursor(NULL,IDC_ARROW);

	This->url = NULL;
	This->mimetype = NULL;
	This->bogusmimetype = 0;
	This->thistype_autodl = 0;
	This->pluginspage = NULL;
	This->tempfilename = NULL;
	This->tempfile = INVALID_HANDLE_VALUE;

	// examine the <embed> tag arguments
	for(i=0;i<argc;i++) {
		if(!_stricmp(argn[i],"type")) {
			This->mimetype = _strdup(argv[i]);
			CheckMIMEType(This);
		}
		else if(!_stricmp(argn[i],"pluginspage")) {
			This->pluginspage = _strdup(argv[i]);
		}
	}
	return NPERR_NO_ERROR;
}

static NPError NPP_Destroy(NPP instance, NPSavedData** save)
{
	PluginInstance* This;

	if (!instance) return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*)instance->pdata;
	if(!This) return NPERR_INVALID_INSTANCE_ERROR;

	// if a tempfile was created, delete it
	if(This->tempfilename) {
		DeleteFile(This->tempfilename);
		free(This->tempfilename); This->tempfilename=NULL;
	}

	if( This->fhWnd ) { /* If we have a window, un-subclass it */
		SetWindowLong( This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc);
		This->fDefaultWindowProc = NULL;
		This->fhWnd = NULL;
	}

	if(This->bkgdbrush1) { DeleteObject(This->bkgdbrush1); This->bkgdbrush1=NULL; }
	if(This->bkgdbrush2) { DeleteObject(This->bkgdbrush2); This->bkgdbrush2=NULL; }
	if(This->bkgdbrush3) { DeleteObject(This->bkgdbrush3); This->bkgdbrush3=NULL; }
	if(This->outlinepen) { DeleteObject(This->outlinepen); This->outlinepen=NULL; }
	if(This->url) { free(This->url); This->url=NULL; }
	if(This->mimetype) { free(This->mimetype); This->mimetype=NULL; }
	if(This->pluginspage) { free(This->pluginspage); This->pluginspage=NULL; }
	if(This->pluginsearchpage) { free(This->pluginsearchpage); This->pluginsearchpage=NULL; }

	if(instance->pdata) { free(instance->pdata); instance->pdata = NULL; }
	return NPERR_NO_ERROR;
}

/* Browser is providing us with a window */
static NPError NPP_SetWindow(NPP instance, NPWindow* window)
{
	NPError result = NPERR_NO_ERROR;
	PluginInstance* This;

	if (!instance) return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;

	if(This->fhWnd) {
		if( (!window) || (!window->window) ) {
			/* There is now no window to use. get rid of the old
			 * one and exit. */
			SetWindowLong( This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc);
			This->fDefaultWindowProc = NULL;
			This->fhWnd = NULL;
			This->fWindow=window;
			return NPERR_NO_ERROR;
		}

		else if ( This->fhWnd == (HWND) window->window ) {
			/* The new window is the same as the old one. Redraw and get out. */
			This->fWindow=window;
			InvalidateRect( This->fhWnd, NULL, FALSE );
			return NPERR_NO_ERROR;
		}
		else {
			/* Unsubclass the old window, so that we can subclass the new
			 * one later. */
			SetWindowLong( This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc);
			This->fDefaultWindowProc = NULL;
			This->fhWnd = NULL;
		}
	}
	else if( (!window) || (!window->window) ) {
		/* We can just get out of here if there is no current
		 * window and there is no new window to use. */
		This->fWindow=window;
		return NPERR_NO_ERROR;
	}

	/* Subclass the new window so that we can begin drawing and
	 * receiving window messages. */
	This->fDefaultWindowProc = (WNDPROC)SetWindowLong( (HWND)window->window, GWL_WNDPROC, (LONG)PluginWindowProc);
	This->fhWnd = (HWND) window->window;
	SetProp( This->fhWnd, gInstanceLookupString, (HANDLE)This);

	This->fWindow = window;
	InvalidateRect( This->fhWnd, NULL, TRUE );
	return result;
}

static NPError NPP_NewStream(NPP instance,NPMIMEType type,NPStream *stream, 
	      NPBool seekable,uint16 *stype)
{
	PluginInstance* This;
	char tmppath[MAX_PATH];

	if(instance==NULL) return NPERR_INVALID_INSTANCE_ERROR;
	if(!stream || !stype) return NPERR_GENERIC_ERROR;
	This = (PluginInstance*) instance->pdata;
	if(!This) return NPERR_GENERIC_ERROR;

	if(This->status>ST_INIT) {
		// An unexpected stream??
		return NPERR_GENERIC_ERROR;
	}

	if(This->tempfilename) {
		// It should be safe to delete the previous file, because
		// we won't get here if the old stream is still loading.
		DeleteFile(This->tempfilename);
		free(This->tempfilename); This->tempfilename=NULL;
	}

	/* save things for for later */
	if(This->url) { free(This->url); This->url=NULL; }
	if(stream->url) This->url = _strdup(stream->url);
	if(type) {
		if(!This->mimetype) {
			This->mimetype = _strdup(type);
			CheckMIMEType(This);
		}
	}

	// NP_ASFILE or NP_ASFILEONLY would be more efficient, but
	// unfortunately they are not completely reliable in some
	// browsers.
	(*stype)=NP_NORMAL;

	This->stream_id = stream;
	This->status=ST_LOADING;
	This->abortflag=0;
	This->bytestotal=stream->end;  // 0 or less means unknown
	This->bytesread=0;

	if(!This->forcedownload) {
		if(This->thistype_autodl==1 || (This->thistype_autodl==0 && This->autodownload==0)) {
			This->status=ST_NOTLOADING;
			RepaintWindow(This);
			return NPERR_GENERIC_ERROR;
		}
	}

	This->tempfilename = malloc(MAX_PATH);
	if(!This->tempfilename) return NPERR_GENERIC_ERROR;

	if(!GetTempPath( MAX_PATH, tmppath)) {
		free(This->tempfilename); This->tempfilename=NULL;
		return NPERR_GENERIC_ERROR;
	}
	if(!GetTempFileName(tmppath,"dpp",0,This->tempfilename)) {
		free(This->tempfilename); This->tempfilename=NULL;
		return NPERR_GENERIC_ERROR;
	}

	This->tempfile = CreateFile(This->tempfilename,GENERIC_WRITE,FILE_SHARE_READ,
	    NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
	if(This->tempfile == INVALID_HANDLE_VALUE) {
		warn(This,"Can't write temporary file");
	}

	RepaintWindow(This);
	return NPERR_NO_ERROR;
}

static int32 NPP_WriteReady(NPP instance, NPStream *stream)
{
	return (int32)0X0FFFFFFF;
}

static int32 NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer)
{
	PluginInstance* This;
	DWORD written;

	if (!instance) return -1;
	This = (PluginInstance*) instance->pdata;
	if (!This) return -1;

	if(stream != This->stream_id) return -1;

	if(This->abortflag) {
		This->status=ST_STOPPED;
		This->abortflag=0;
		RepaintWindow(This);
		return -1;
	}

	if(This->status==ST_NOTLOADING || This->status==ST_STOPPED) {
		return -1;
	}

	if(This->tempfile == INVALID_HANDLE_VALUE) return -1;

	if(len>0) {
		WriteFile(This->tempfile,buffer,len,&written,NULL);
		This->bytesread += len;

		RepaintWindow(This);
	}

	/* Return the number of bytes accepted -- we always accept them all. */
	return len; 
}

/* DestroyStream gets called after the file has finished loading */
static NPError NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
	PluginInstance* This;

	if (!instance) return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*) instance->pdata;
	if (!This) return NPERR_INVALID_INSTANCE_ERROR;

	if(stream != This->stream_id) {
		return NPERR_NO_ERROR; // not the stream we expected
	}

	This->stream_id = NULL;

	if(This->tempfile != INVALID_HANDLE_VALUE) {
		CloseHandle(This->tempfile);
		This->tempfile = INVALID_HANDLE_VALUE;
	}

	if(reason==NPRES_DONE && This->status==ST_LOADING) {
		This->status=ST_LOADED;
		This->bytestotal = This->bytesread;
	}
	else {
		if(This->status==ST_LOADING) This->status=ST_STOPPED;
		This->abortflag=0;
	}

	RepaintWindow(This);
	return NPERR_NO_ERROR;
}

/* This should not get called, since we aren't using the NP_ASFILE mode */
static void NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
	return;
}

/* Print embedded plug-in (via the browser's Print command) */
static void NPP_Print(NPP instance, NPPrint* printInfo)
{
	return;
}

static void NPP_URLNotify(NPP instance, const char* url, NPReason reason, void* notifyData)
{
	return;
}

/**********************************************************************/

static int CopyToClipboard(PluginInstance *This,unsigned char *mem,int size,UINT format)
{
	HGLOBAL hClip;
	LPVOID lpClip;
	int ok = 0;

	if(!mem) return 0;
	if(!OpenClipboard(NULL)) return 0;

	if(EmptyClipboard()) {
		hClip=GlobalAlloc(GMEM_ZEROINIT|GMEM_MOVEABLE|GMEM_DDESHARE,size);
		lpClip=GlobalLock(hClip);
		if(lpClip) {
			CopyMemory(lpClip,mem,size);
			GlobalUnlock(hClip);
			if(SetClipboardData(format,hClip)) ok=1;
		}
		if(!ok) GlobalFree(hClip);
	}
	CloseClipboard();
	return ok;
}

static int SaveToFile(PluginInstance *This, char *filename)
{
	if(!CopyFile(This->tempfilename,filename,FALSE)) {
		warn(This,"Can't save file");
	}
	return 1;
}

/* Try to make a filename from the url. Caller must provide fn[MAX_PATH]
 * buffer.
 * This function attempts to extract a bitmap filename from a URL,
 * but if it doesn't look like it contains an appropriate name,
 * it leaves it blank.
 * FIXME: This could be better. Should at least convert %20 to space.
 */
static void url2filename(char *fn, char *url)
{
	int title,ext,i;

	lstrcpy(fn,"");
	ext=0;   /* position of the file extention */
	title=0; /* position of the base filename */
	for(i=0;url[i];i++) {
		if(url[i]=='.') ext=i+1;
		if(url[i]=='/') title=i+1;
		if(url[i]=='\\') title=i+1;
		if(url[i]==':') title=i+1;
		if(url[i]=='=') title=i+1;
		if(url[i]=='?') title=i+1;
	}

	lstrcpyn(fn,&url[title],MAX_PATH);
}

// Write the document to a local file 
static void SaveDocument(PluginInstance *This)
{
	OPENFILENAME ofn;
	char fn[MAX_PATH];

	if(This->status!=ST_LOADED || This->tempfilename==NULL) {
		warn(This,"File not loaded -- can't save");
		return;
	}

	if(lstrlen(This->url)) {
		url2filename(fn,This->url);
	}
	else {
		lstrcpy(fn,"");
	}

	ZeroMemory(&ofn,sizeof(OPENFILENAME));
	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=This->fhWnd;
	ofn.lpstrFilter="*.*\0*.*\0\0";
	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Save As...";
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_PATHMUSTEXIST|OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT;
	ofn.lpstrDefExt=NULL;

	if(GetSaveFileName(&ofn)) {
		SaveToFile(This,ofn.lpstrFile);
	}
}

static void PluginSearch(PluginInstance *This)
{
	char *searchpage;
	char searchurl[1000];

	if(!This->mimetype) return;
	if(This->pluginsearchpage)
		searchpage = This->pluginsearchpage;
	else
		searchpage = g_DefPluginSearchPage;

	// FIXME: should URLencode the mime type
	_snprintf(searchurl,1000,"%s?mimetype=%s",searchpage,This->mimetype);

	NPN_GetURL(This->instance,searchurl,"_blank");
}

static void ContextMenu(PluginInstance *This, HWND hwnd)
{
	HMENU menu,submenu1;
	int cmd;
	POINT pt;

	pt.x=pt.y=0;
	GetCursorPos(&pt);

	submenu1=CreatePopupMenu();
	AppendMenu(submenu1,MF_ENABLED|(This->thistype_autodl==2?MF_CHECKED:0),ID_AUTODLYES,"&Yes");
	AppendMenu(submenu1,MF_ENABLED|(This->thistype_autodl==1?MF_CHECKED:0),ID_AUTODLNO,"&No");
	AppendMenu(submenu1,MF_ENABLED|(This->thistype_autodl==0?MF_CHECKED:0),ID_AUTODLDEF,"Use &Default");

	menu=CreatePopupMenu();
	AppendMenu(menu,(This->url && This->status==ST_LOADED)?MF_ENABLED:MF_GRAYED,ID_SAVEAS,"&Save Document...");
	AppendMenu(menu,This->url?MF_ENABLED:MF_GRAYED,ID_VIEWDIRECT,"&View Document Directly");
	AppendMenu(menu,This->url?MF_ENABLED:MF_GRAYED,ID_OPENNEWWIN,"&Open in New Window");
	AppendMenu(menu,This->url?MF_ENABLED:MF_GRAYED,ID_COPYURL,"Cop&y Document Location");
	AppendMenu(menu,(This->pluginspage)?MF_ENABLED:MF_GRAYED,ID_PLUGINSPAGE,"Get Recommended Plug-in");
	AppendMenu(menu,(This->mimetype && !This->bogusmimetype)?MF_ENABLED:MF_GRAYED,ID_PLUGINSEARCH,"Plug-in Search");
	AppendMenu(menu,MF_ENABLED,ID_REFRESHPLUGINS,"Refresh Plug-ins");
	AppendMenu(menu,MF_SEPARATOR,0,NULL);
	AppendMenu(menu,(This->url && This->status!=ST_LOADING && This->status!=ST_LOADED)?MF_ENABLED:MF_GRAYED,
		ID_LOADDOC,"&Download Document");
	AppendMenu(menu,(This->status==ST_LOADING)?MF_ENABLED:MF_GRAYED,
		ID_STOPLOAD,"S&top Downloading");
	AppendMenu(menu,MF_SEPARATOR,0,NULL);
	AppendMenu(menu,MF_POPUP|(This->mimetype?MF_ENABLED:MF_GRAYED),(UINT)submenu1,"AutoDownload This Type");
	AppendMenu(menu,MF_ENABLED|(This->autodownload?MF_CHECKED:0),ID_AUTODOWNLOAD,"AutoDownload By Default");
	AppendMenu(menu,MF_SEPARATOR,0,NULL);
	AppendMenu(menu,MF_ENABLED,ID_PROPERTIES,"P&roperties...");
	AppendMenu(menu,MF_SEPARATOR,0,NULL);
	AppendMenu(menu,MF_ENABLED,ID_ABOUT,"About DefaultPluginPro...");

	cmd=TrackPopupMenuEx(menu, TPM_LEFTALIGN|TPM_TOPALIGN|TPM_NONOTIFY|TPM_RETURNCMD|
		TPM_RIGHTBUTTON,pt.x,pt.y,hwnd,NULL);

	DestroyMenu(menu);
	DestroyMenu(submenu1);

	switch(cmd) {
	case ID_SAVEAS:
		SaveDocument(This);
		break;
	case ID_LOADDOC:
		if(This->url) {
			This->status=ST_INIT;
			This->forcedownload=1;
			RepaintWindow(This);
			NPN_GetURL(This->instance,This->url,NULL);
		}
		break;
	case ID_STOPLOAD:
		if(This->status==ST_LOADING) {
			This->abortflag=1;
		}
		break;
	case ID_COPYURL:
		if(lstrlen(This->url)) {
			CopyToClipboard(This,This->url,lstrlen(This->url)+1,CF_TEXT);
		}
		else {
			warn(This,"No link to copy");
		}
		break;
	case ID_VIEWDIRECT:
		if(This->url) NPN_GetURL(This->instance,This->url,"_self");
		break;
	case ID_OPENNEWWIN:
		if(This->url) NPN_GetURL(This->instance,This->url,"_blank");
		break;
	case ID_PLUGINSPAGE:
		if(This->pluginspage)
			NPN_GetURL(This->instance,This->pluginspage,"_blank");
		break;
	case ID_PLUGINSEARCH:
		PluginSearch(This);
		break;
	case ID_REFRESHPLUGINS:
		NPN_GetURL(This->instance,"javascript:navigator.plugins.refresh(true)","_self");
		break;
	case ID_PROPERTIES:
		DialogBoxParam(g_hInst,"PROPERTIESDLG",This->fhWnd,(DLGPROC)DlgProcProperties,(LPARAM)This);
		break;
	case ID_ABOUT:
		DialogBoxParam(g_hInst,"ABOUTDLG",This->fhWnd,(DLGPROC)DlgProcAbout,(LPARAM)This);
		break;
	case ID_AUTODOWNLOAD:
		This->autodownload = !This->autodownload;
		WriteSetting_Int(This,NULL,"autodownload",This->autodownload);
		break;
	case ID_AUTODLYES:
		This->thistype_autodl=2;
		goto here;
	case ID_AUTODLNO:
		This->thistype_autodl=1;
		goto here;
	case ID_AUTODLDEF:
		This->thistype_autodl=0;
here:
		WriteSetting_Int(This,"types",This->mimetype,This->thistype_autodl);
		break;
	}
}

static int PaintProc(PluginInstance *This, HWND hWnd, HDC hdc)
{
	char buf[1000];
	RECT rect,r;
	int textheight,voffset,hoffset;
	TEXTMETRIC tm;
	char *Url;
	int i;

	Url = This->url;
	if(Url) {
		for(i=0;This->url[i];i++) {
			if(This->url[i]=='/' || This->url[i]=='\\') {
				if(This->url[i+1]) {
					Url = &This->url[i+1];
				}
			}
		}
	}

	if(GetTextMetrics(hdc,&tm)) {
		textheight = tm.tmHeight;
	}
	else {
		textheight=18;
	}
	voffset=hoffset=2;

	GetClientRect(hWnd,&rect);

	i=0;
	if(This->status==ST_LOADED) {
		i=rect.right;
	}
	else if(This->status==ST_LOADING) {
		if(This->bytestotal>0) {
			i =  (int) ( ((double)This->bytesread/This->bytestotal) * (double)rect.right);
		}
	}

	r = rect; // structure copy

	if(This->status==ST_LOADED) {
		FillRect(hdc,&r,This->bkgdbrush3);
	}
	else {
		r.right = i;
		FillRect(hdc,&r,This->bkgdbrush2);
		r.left=i; r.right=rect.right;
		FillRect(hdc,&r,This->bkgdbrush1);
	}

	SetBkMode(hdc,TRANSPARENT);
	SelectObject(hdc,GetStockObject(NULL_BRUSH));
	SelectObject(hdc,This->outlinepen);
	Rectangle(hdc,rect.left,rect.top,rect.right,rect.bottom);

	r.left=hoffset; r.right=rect.right-hoffset;  // RECT for text line

	if(rect.bottom>7*textheight && rect.right>120) {
		if(This->status==ST_LOADED)
			SetTextColor(hdc,RGB(146,224,146));
		else
			SetTextColor(hdc,RGB(214,165,108));
			
		DrawText(hdc,"<<< DefaultPluginPro >>>",-1,&rect,DT_CENTER|DT_VCENTER|DT_SINGLELINE);
	}

	if(Url) {
		r.top=voffset; r.bottom=voffset+textheight;
		SetTextColor(hdc,RGB(0,0,0));
		DrawText(hdc,Url,-1,&r,DT_LEFT|DT_NOPREFIX|DT_SINGLELINE|DT_END_ELLIPSIS);
	}
	if(This->mimetype) {
		r.top=voffset+textheight; r.bottom=voffset+2*textheight;
		SetTextColor(hdc,RGB(64,64,192));
		DrawText(hdc,This->mimetype,-1,&r,DT_LEFT|DT_NOPREFIX|DT_SINGLELINE|DT_END_ELLIPSIS);
	}

	if(This->status==ST_INIT) {
		sprintf(buf,"Waiting for data");
	}
	else if(This->status==ST_NOTLOADING) {
		sprintf(buf,"Not downloaded");
	}
	else if(This->status==ST_STOPPED) {
		sprintf(buf,"Download aborted");
	}
	else if(This->status==ST_LOADING) {
		if(This->bytestotal>0) {
			sprintf(buf,"Loading... %d/%d bytes",This->bytesread,This->bytestotal);
		}
		else {
			sprintf(buf,"Loading... %d bytes",This->bytesread);
		}
	}
	else if(This->status==ST_LOADED) {
		sprintf(buf,"Download complete (%d bytes)",This->bytestotal);
	}
	else {
		sprintf(buf,"Error");
	}

	r.top=voffset+2*textheight; r.bottom=voffset+3*textheight;
	SetTextColor(hdc,RGB(128,64,64));
	DrawText(hdc,buf,-1,&r,DT_LEFT|DT_NOPREFIX|DT_SINGLELINE|DT_END_ELLIPSIS);

	return 1;
}

static LRESULT CALLBACK PluginWindowProc( HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{

	PluginInstance* This = (PluginInstance*) GetProp(hWnd, gInstanceLookupString);
	if(!This) return DefWindowProc( hWnd, Msg, wParam, lParam);

	switch(Msg) {
	case WM_ERASEBKGND:
		return 1;

	case WM_CONTEXTMENU: case WM_RBUTTONUP:
		ContextMenu(This,/* LOWORD(lParam),HIWORD(lParam), */ hWnd);
		return 0;

	case WM_MOUSEMOVE:
		SetCursor(This->cursor);
		return 0;

	case WM_PAINT:
		{
			PAINTSTRUCT paintStruct;
			HDC hdc;

			hdc = BeginPaint( hWnd, &paintStruct );

			PaintProc(This,hWnd,hdc);


			EndPaint( hWnd, &paintStruct );
			return 0;
		}
	}
	/* Forward unprocessed messages on to their original destination
	 * (the window proc we replaced) */
	return This->fDefaultWindowProc(hWnd, Msg, wParam, lParam);
}

static LRESULT CALLBACK DlgProcAbout(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	switch(Msg) {
	case WM_INITDIALOG:
		{
			char buf[4096];

			PluginInstance *This=(PluginInstance*)lParam;

			sprintf(buf,"DefaultPluginPro: A default plug-in replacement for web browsers.\r\n\r\n"
				"Version %s\r\n%s"
#ifdef _DEBUG
				" DEBUG BUILD"
#endif
				"\r\nCopyright 2002 by Jason Summers\r\n"
				"http://pobox.com/~jason1/defaultpluginpro/\r\n\r\n"
				"This software may be distributed only under the terms of the GNU "
				"General Public License, version 2 or later. See the file readme.txt "
				"for more information."
				,DPPVERS,__DATE__);
			SetDlgItemText(hWnd,IDC_PRGINFO,buf);

		}
		return(TRUE);
	case WM_CLOSE:
		EndDialog(hWnd,0);
		return(TRUE);
	case WM_COMMAND:
		switch(wParam) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hWnd,0);
			return(TRUE);
		}
	}
	return(FALSE);
}

static LRESULT CALLBACK DlgProcProperties(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	switch(Msg) {
	case WM_INITDIALOG:
		{
			char buf[8192];
			char info[1024];
			char buf2[512];
			DWORD tabs[1];

			PluginInstance *This=(PluginInstance*)lParam;

			tabs[0]= 80;
			SendDlgItemMessage(hWnd,IDC_DOCINFO,EM_SETTABSTOPS,(WPARAM)1,(LPARAM)tabs);
			lstrcpy(buf,"");

			if(This->url) {
				_snprintf(info,200,"Document URL:\t%s\r\n",This->url);
				lstrcat(buf,info);
			}

			if(This->mimetype) {
				_snprintf(info,200,"Type:\t%s\r\n",This->mimetype);
				lstrcat(buf,info);
			}

			if(This->bytestotal>0) {
				sprintf(info,"Size:\t%d bytes\r\n",This->bytestotal);
				lstrcat(buf,info);
			}

			if(This->pluginspage) {
				_snprintf(info,200,"Plug-in web page:\t%s\r\n",This->pluginspage);
				lstrcat(buf,info);
			}

			if(This->tempfilename) {
				sprintf(info,"Temp file:\t%s\r\n",This->tempfilename);
				lstrcat(buf,info);
			}

			if(GetModuleFileName(g_hInst,buf2,512)) {
				sprintf(info,"DefPluginPro location:\t%s",buf2);
				lstrcat(buf,info);
			}

			SetDlgItemText(hWnd,IDC_DOCINFO,buf);
		}
		return(TRUE);
	case WM_CLOSE:
		EndDialog(hWnd,0);
		return(TRUE);
	case WM_COMMAND:
		switch(wParam) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hWnd,0);
			return(TRUE);
		}
	}
	return(FALSE);
}

/////////////////////////
/////////////////////////  exported functions, etc.

static JRIGlobalRef Private_GetJavaClass(void)
{
    return NULL;
}

NPError WINAPI NP_GetEntryPoints(NPPluginFuncs* pFuncs)
{
    if(!pFuncs) return NPERR_INVALID_FUNCTABLE_ERROR;

    pFuncs->version       = (NP_VERSION_MAJOR << 8) | NP_VERSION_MINOR;
    pFuncs->newp          = NPP_New;
    pFuncs->destroy       = NPP_Destroy;
    pFuncs->setwindow     = NPP_SetWindow;
    pFuncs->newstream     = NPP_NewStream;
    pFuncs->destroystream = NPP_DestroyStream;
    pFuncs->asfile        = NPP_StreamAsFile;
    pFuncs->writeready    = NPP_WriteReady;
    pFuncs->write         = NPP_Write;
    pFuncs->print         = NPP_Print;
    pFuncs->event         = NULL;

	g_pluginFuncs		  = pFuncs;

    return NPERR_NO_ERROR;
}

NPError WINAPI NP_Initialize(NPNetscapeFuncs* pFuncs)
{
	int navMinorVers;

    if(!pFuncs) return NPERR_INVALID_FUNCTABLE_ERROR;

    g_pNavigatorFuncs = pFuncs; // save it for future reference 

    if(HIBYTE(pFuncs->version) > NP_VERSION_MAJOR)
        return NPERR_INCOMPATIBLE_VERSION_ERROR;

    navMinorVers = g_pNavigatorFuncs->version & 0xFF;
	if(navMinorVers>=NPVERS_HAS_NOTIFICATION)
		g_pluginFuncs->urlnotify = NPP_URLNotify;
	if( navMinorVers>=NPVERS_HAS_LIVECONNECT)
		g_pluginFuncs->javaClass = Private_GetJavaClass();

	// custom stuff could go here
    //return NPP_Initialize();

	return NPERR_NO_ERROR;
}

NPError WINAPI NP_Shutdown()
{
	// custom stuff could go here
    //NPP_Shutdown();

    g_pNavigatorFuncs = NULL;
    return NPERR_NO_ERROR;
}

char* WINAPI NP_GetMIMEDescription()
{
  static char mimetype[] = "*";
  return mimetype;
}

BOOL APIENTRY DllMain(HANDLE hModule, DWORD reason, LPVOID lpReserved)
{
    switch(reason) {
    case DLL_PROCESS_ATTACH:
		g_hInst=hModule;
		break;
    case DLL_THREAD_ATTACH:
		break;
    case DLL_THREAD_DETACH:
		break;
    case DLL_PROCESS_DETACH:
		break;
    }
    return TRUE;
}
